/*
   (c) Copyright 2000-2002  convergence integrated media GmbH.
   All rights reserved.

   Written by Denis Oliver Kropp <dok@directfb.org>,
              Andreas Hundt <andi@fischlustig.de> and
              Sven Neumann <neo@directfb.org>.
              
   This file is subject to the terms and conditions of the MIT License:

   Permission is hereby granted, free of charge, to any person
   obtaining a copy of this software and associated documentation
   files (the "Software"), to deal in the Software without restriction,
   including without limitation the rights to use, copy, modify, merge,
   publish, distribute, sublicense, and/or sell copies of the Software,
   and to permit persons to whom the Software is furnished to do so,
   subject to the following conditions:

   The above copyright notice and this permission notice shall be
   included in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
   CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
   TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
   SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stdio.h>

#include <directfb.h>

/* the super interface */
IDirectFB *dfb;

/* the primary surface (surface of primary layer) */
IDirectFBSurface *primary;

/* provider for our images/font */
IDirectFBFont               *font;

/* Input interfaces: device and its buffer */
IDirectFBInputDevice *keyboard;
IDirectFBEventBuffer *key_events;

/* macro for a safe call to DirectFB functions */
#define DFBCHECK(x...) \
        {                                                                      \
           err = x;                                                            \
           if (err != DFB_OK) {                                                \
              fprintf( stderr, "%s <%d>:\n\t", __FILE__, __LINE__ );           \
              DirectFBErrorFatal( #x, err );                                   \
           }                                                                   \
        }

static char *rules[] = { "CLEAR", "SRC", "SRC OVER", "DST OVER",
                         "SRC IN", "DST IN", "SRC OUT", "DST OUT" };
static int num_rules = sizeof( rules ) / sizeof( rules[0] );

static int screen_width, screen_height;

int main( int argc, char *argv[] )
{
     int i;
     DFBResult err;

     DFBCHECK(DirectFBInit( &argc, &argv ));

     /* create the super interface */
     DFBCHECK(DirectFBCreate( &dfb ));

     /* get an interface to the primary keyboard
        and create an input buffer for it */
     DFBCHECK(dfb->GetInputDevice( dfb, DIDID_KEYBOARD, &keyboard ));
     DFBCHECK(keyboard->CreateEventBuffer( keyboard, &key_events ));

     /* set our cooperative level to DFSCL_FULLSCREEN
        for exclusive access to the primary layer */
     dfb->SetCooperativeLevel( dfb, DFSCL_FULLSCREEN );

     /* get the primary surface, i.e. the surface of the
        primary layer we have exclusive access to */
     {
          DFBSurfaceDescription dsc;

          dsc.flags = DSDESC_CAPS | DSDESC_PIXELFORMAT;
          dsc.caps = DSCAPS_PRIMARY | DSCAPS_VIDEOONLY;
          dsc.pixelformat = DSPF_ARGB;

          DFBCHECK(dfb->CreateSurface( dfb, &dsc, &primary ));
     }

     DFBCHECK(primary->GetSize( primary, &screen_width, &screen_height ));

     primary->SetColor( primary, 30, 40, 50, 50 );
     DFBCHECK(primary->FillRectangle( primary, 0, 0,
                                      screen_width, screen_height ));

     {
          DFBFontDescription desc;

          desc.flags = DFDESC_HEIGHT;

          desc.height = screen_width/32;
          DFBCHECK(dfb->CreateFont( dfb, FONT, &desc, &font ));
          DFBCHECK(primary->SetFont( primary, font ));
          DFBCHECK(primary->SetColor( primary, 0xFF, 0xFF, 0xFF, 0xFF ));
          DFBCHECK(primary->DrawString( primary, "Porter/Duff Demo", -1,
                                        screen_width/2, 50, DSTF_CENTER ));
          DFBCHECK(font->Release( font ));

          desc.height = screen_width/42;
          DFBCHECK(dfb->CreateFont( dfb, FONT, &desc, &font ));
          DFBCHECK(primary->SetFont( primary, font ));
     }

     for (i=0; i<num_rules; i++) {
          DFBCHECK(primary->SetDrawingFlags( primary, DSDRAW_NOFX ));
          DFBCHECK(primary->SetColor( primary, 50, 250, 50, 200 ));
          DFBCHECK(primary->FillRectangle( primary, i*(screen_width-40)/num_rules + 40,
                                           130, 50, 70 ));

          DFBCHECK(primary->SetPorterDuff( primary, i+1 ));
          DFBCHECK(primary->SetColor( primary, 250, 50, 50, 100 ));
          DFBCHECK(primary->SetDrawingFlags( primary, DSDRAW_BLEND ));
          DFBCHECK(primary->FillRectangle( primary, i*(screen_width-40)/num_rules + 50,
                                           140, 50, 70 ));

          DFBCHECK(primary->SetPorterDuff( primary, DSPD_NONE ));
          DFBCHECK(primary->SetColor( primary, i*0x1F, i*0x10+0x7f, 0xFF, 0xFF ));
          DFBCHECK(primary->DrawString( primary, rules[i], -1,
                                        i*(screen_width-40)/num_rules + 65, 220,
                                        DSTF_CENTER | DSTF_TOP ));
     }

     key_events->Reset( key_events );

     while (1) {
          DFBInputEvent ev;

          key_events->WaitForEvent( key_events );

          key_events->GetEvent( key_events, DFB_EVENT(&ev) );

          if (ev.type == DIET_KEYRELEASE && ev.key_symbol == DIKS_ESCAPE)
               break;
     }

     /* release our interfaces to shutdown DirectFB */
     primary->Release( primary );
     font->Release( font );
     key_events->Release( key_events );
     keyboard->Release( keyboard );
     dfb->Release( dfb );

     return 0;
}

